#include <stdio.h>
#include <string.h>
#include "f320.h"

// Nokia 3310 LCD driver code modified from the following sources :
// http://www.geocities.com/dariuskrail20/nokia3310lcd.html
// http://www.microsyl.com/ 
// I changed the x2size font printing code to avoid using a RAM screen cache.
// Flash read/write source code modified from Silicon Labs app note AN201. As I was running
// short of code space, I modified their code to use a RAM scratch buffer 
// instead of a flash scratch page, and simplified it to assume writes & reads do 
// not cross a page boundary.
// SMBus (i2c) code cosmetically modified from SiLabs example code


//#define PROFILE
//#define FLOG_USE_SERIAL_EEPROM
//#define VARIO_DAMPING
#define FLOG_USE_FLASH


typedef unsigned long  u32;
typedef unsigned short u16;
typedef unsigned char  u8;
typedef signed long    s32;
typedef signed short   s16;
typedef signed char    s8;

#define ABS(x)                 ((x) < 0 ? -(x) : (x))
#define CLAMP(x,min,max)       {if ((x) <= (min)) (x) = (min); else if ((x) >= (max)) (x) = (max);}
#define CORE(x,t)              {if (ABS(x) <= (t)) (x) = 0;}



#define MIN_ALTITUDE_M          -999
#define MAX_ALTITUDE_M          9999

#define MIN_TEMP_C               -99
#define MAX_TEMP_C               999

#define MIN_CLIMBRATE_CPS        -2500
#define MAX_CLIMBRATE_CPS         2500

#define BEEP_DISCR_THRESHOLD   20

#define CLIMBTONE_CPS_MIN   0
#define CLIMBTONE_CPS_DEF   0
#define CLIMBTONE_CPS_MAX   75

#define SINKTONE_CPS_MIN   100
#define SINKTONE_CPS_DEF   300
#define SINKTONE_CPS_MAX   400

#define PZLUT_ENTRIES          56
#define PA_INIT                102860L
#define PA_DELTA               1024L

#define INTERRUPT_INT0             0   // External Interrupt 0
#define INTERRUPT_TIMER0           1   // Timer0 Overflow
#define INTERRUPT_INT1             2   // External Interrupt 1
#define INTERRUPT_TIMER1           3   // Timer1 Overflow
#define INTERRUPT_UART0            4   // Serial Port 0
#define INTERRUPT_TIMER2           5   // Timer2 Overflow
#define INTERRUPT_SPI0             6   // Serial Peripheral Interface 0
#define INTERRUPT_SMBUS0           7   // SMBus0 Interface
#define INTERRUPT_ADC0             10  // ADC0 conversion complete 
#define INTERRUPT_TIMER3           14  // Timer3 Overflow

// SM bus status  - top 4 bits only
#define SMB_MTSTA      0xE0           // (MT) start transmitted
#define SMB_MTDB       0xC0           // (MT) data byte transmitted
#define SMB_MRDB       0x80           // (MR) data byte received
#define SMB_WRITE      0x00           // SMBus write command
#define SMB_READ       0x01           // SMBus read command
#define SMB_BUFF_SIZE  0x08           // Defines the maximum number of bytes

//#define SMB_ID_24LC02          ((u8)0xA0)
#define SMB_ID_DS1338          ((u8)0xD0)

#ifdef FLOG_USE_FLASH
   #define FLASH_PAGESIZE        512
   #define FLOG_PAGE_ADDR       0x03A00L // last usable flash page
   #define FLOG_MAX_FLIGHTS      30

   void flash_WriteByte (u16 faddr, char byte);
   u8   flash_ReadByte (u16 faddr);
   void flash_ErasePage (u16 faddr);
   void flash_WriteBuf (u16 faddr, char *pBuf, u16 nbytes);
   char * flash_ReadBuf (char *pBuf, u16 faddr, u16 nbytes);
   void flash_Clear (u16 faddr, u16 nbytes);
   void flash_Update (u16 faddr, char *pBuf, u16 nbytes);
   void flash_Fill (u16 faddr, u16 length, u8 val);
   void flash_EraseScratchBuf(void);  
#endif

#ifdef FLOG_USE_SERIAL_EEPROM
   #define FLOG_MAX_FLIGHTS     14
   #define FLOG_EEP_ADDR        10
#endif

#define FLOG_THRESHOLD_M     5
#define FLOG_REC_BYTES       17

#define NVD_ADDR_FLOG_OLDEST_REC 8
#define NVD_ADDR_PZTBL_SEL       9
#define NVD_ADDR_SINKTONE_CPS    10
#define NVD_ADDR_CLIMBTONE_CPS   11
#define NVD_ADDR_CKSUM           12
 
#define UIMODE_FLOG   11
#define UIMODE_SET    22
#define UIMODE_RUN    33
#define UIMODE_SLEEP  44

#define RTC_ADDR_SEC    0
#define RTC_ADDR_MIN    1
#define RTC_ADDR_HRS    2
#define RTC_ADDR_DOW    3
#define RTC_ADDR_DOM    4
#define RTC_ADDR_MON    5
#define RTC_ADDR_YAR    6
#define RTC_AM      0
#define RTC_PM      1
#define RTC_HRS_24  0
#define RTC_HRS_12  1

#define SEL_ALTITUDE  0
#define SEL_SINKTONE  1
#define SEL_CLIMBTONE 2
#define SEL_RTC_HRS   3
#define SEL_RTC_MIN   4
#define SEL_RTC_DOM   5
#define SEL_RTC_MON   6
#define SEL_RTC_YAR   7

// slightly less than 70mS per {Pressure,Temperature} sample
#define SNS_ACQ_PER_SEC         15
#define SNS_NUM_Z_SAMPLES       33

volatile u8  gnHrs;
volatile u8  gnMin;
volatile u8  gnSec;
u8  gnDom;
u8  gnYar;
u8  gnMon;

u8* gpSmbDataRcv;           // Global pointer for SMBus data All receive data is written here
u8 gnSmbData;               // Global holder for single byte writes.
u8* gpSmbDataXmt;           // Global pointer for SMBus data. All transmit data is read from here
u8 gnSmbDataLen;            // Global holder for number of bytes to send or receive 
u8 gnSmbAddr;               // Global holder for the EEPROM word address that will be accessed in
u8 gnSmbDevId;              // Target SMBus slave address

volatile bit gbSmbBusy = 0;          // Software flag to indicate when the access functions have claimed the SMBus
volatile bit gbSmbRW;                // Software flag to indicate the direction of the current transfer
volatile bit gbSmbSendAddr;          // When set, this flag causes the ISR to send the 8-bit <gnSmbAddr> after sending the slave address.
volatile bit gbSmbRandomRead;        // When set, this flag causes the ISR to send a start signal after sending the word address.
volatile bit gbSmbAckPoll;           // When set, this flag causes the ISR to send a repeated start until the slave has acknowledged its address

bit gbHrsMode;
bit gbAmPm;

bit gbFlightLog = 0;
bit gbStable = 0; 

sbit gpioSCLK    = P1^0;					
sbit gpioSDA     = P0^6;
sbit gpioDC	     = P0^7;
sbit gpioExtClk  = P0^3;

sbit UartTx	     = P0^4;
sbit UartRx      = P0^5;
sbit gpioRST_    = P0^2;	

sbit gpioI2CSDA  = P1^3;                       
sbit gpioI2CSCL  = P1^4;                       

sbit gpioBtnR    = P1^1;
sbit gpioBtnL    = P2^5;
sbit gpioBtnS    = P2^7;
sbit gpioBtnP    = P0^1;

sbit gpioMSCLK   = P2^2;   
sbit gpioMDOUT1  = P2^1;
sbit gpioMDIN    = P2^3;

sbit gpioMDOUT2  = P2^4;

xdata u8 gnLCDX;
xdata u8 gnLCDY;

xdata u16 gnBatLifeMin = 0;

xdata u16 gwD1[2];
xdata u16 gwD2[2];
xdata u16 gwC1[2];
xdata u16 gwC2[2];
xdata u16 gwC3[2];
xdata u16 gwC4[2];
xdata u16 gwC5[2];
xdata u16 gwC6[2];

xdata s32 gP[2];
xdata s16 gT[2];

xdata u8 gnDispCnt;

xdata s16 gnSinkToneCps = SINKTONE_CPS_DEF;
xdata s16 gnClimbToneCps = CLIMBTONE_CPS_DEF;

xdata u8 gnUIMode;
xdata u8 gnParSel;
xdata u8 gnAcqState = 0;
xdata u8 gnPZTblSel = 1;

xdata s32 gZBuf[SNS_NUM_Z_SAMPLES];
xdata u8  gnSmpCnt;
xdata s16 gnAvgAltM; 
xdata s32 gnAltCm;
xdata s16 gnCps;
xdata s32 gnCpsx32;
xdata s16 gnTempCx10;

volatile s16 gnBeepPeriodTicks = 0;
volatile s16 gnBeepTransition;
volatile u8  gnBeepFreqReg;
volatile s16 gnBeepCps = 0;
volatile s16 gnVarioCps = 0;

xdata s16 gnMaxAltM;
xdata s16 gnMaxCps;
xdata s16 gnMinCps;
xdata s16 gnMaxTempC;
xdata s16 gnMinTempC;
xdata s16 gnTempC;
xdata s16 gnLaunchAltM;
xdata s16 gnLZAltM;
xdata u8  gnLaunchHrs;
xdata u8  gnLaunchMin;

xdata u8 gnFLogOldestRec; // memory  index
xdata u8 gnFLogRec;    // display index
xdata u8 gnFLBuf[17];
xdata u8 gszBuf[15];

#ifdef FLOG_USE_FLASH
xdata u8 gScratchBuf[FLASH_PAGESIZE];
#endif

xdata u8 gnVBatx10;

volatile s32 gnTimeTick = 0;
volatile u8 gnSecTick = 0;
volatile u8 gnFlightHrs = 0;
volatile u8 gnFlightMin = 0;
volatile u8 gnFlightSec = 0;
volatile u8 gBtnLState = 0;
volatile u8 gBtnRState = 0;
volatile u8 gBtnSState = 0;
volatile u8 gBtnPState = 0;
volatile bit gbBtnLPressed = 0;
volatile bit gbBtnRPressed = 0;
volatile bit gbBtnPPressed = 0;
volatile bit gbBtnSPressed = 0;
volatile bit gbTmr2Flag = 0;

sfr16 ADC0     = 0xbd;                 // ADC0 result

void bat_Voltage(void);

void tmr_WaitTicks(u8 ticks);

void smb_wbyte(u8 devId, u8 addr, u8 dat);
u8   smb_rbyte(u8 devId, u8 addr);
void smb_rbuf(u8 devId, u8* pBuf, u8 addr, u8 len);


void sns_Init(void);
void sns_Reset();
void sns_ReadCoeff(u8 addr, u16* pc1, u16* pc2);
void sns_ReadWord(u16* pw1, u16* pw2);
void sns_SendCmd(u32 cmd, u8 nbits);
void sns_ReadCoeffs();
void sns_TriggerPressureSample();
void sns_TriggerTemperatureSample();
void sns_ConvertPressureTemperature(u8 sel);
void sns_ReadSampleContinuous(void);
void sns_ReadSample(void);
void sns_InitZBuf(void);

s32  var_Pa2Cm(s32 pa);
void var_ClimbRate();
#ifdef VARIO_DAMPING
#define KLPF                   25L
void var_LPFClimbRate(void);
#endif
void var_AverageAltitude();
void var_writeFlightLog(void);
void var_updateFlightData(void);
void var_displayFlightData(void);

void ui_PrintAltitude(char* szBuf, s16 nAlt);
void ui_PrintAltitudeAL(char* szBuf, s16 nAlt, s16 nLaunchAlt);
void ui_PrintTemperature(char* szBuf, s16 nTemp);
void ui_PrintClimbRate(char* szBuf, s16 nCps);
void ui_PrintRealTime(char* szBuf);

void ui_noLZU(char* szBuf);
void ui_noLZS(char* szBuf);

void ui_DisplaySettings(void);
void ui_SetEventHandler(void);
void ui_RunEventHandler(void);
void ui_FLogEventHandler(void);
void ui_SleepHandler(void);
void ui_InitRunMode(void);
void ui_InitSetMode(void);
void ui_InitSleepMode(void);
void ui_Sleep(void);
void ui_InitFLogMode(void);
void ui_DisplayFLog(void);

void mcu_Init(void);

void lcd_Init(void);				
void lcd_WriteCmd(char cmd);			
void lcd_WriteDat(char dat);			
void lcd_ClockData(char dat);				
void lcd_ClearScreen(void);			
void lcd_SetPos(char y,char x);			
void lcd_PutChar(char ch);					
void lcd_PrintSz(char * sz);			
void lcd_PutCharX2 (char ch);
void lcd_PrintSzX2(char *sz);

void rtc_Init(void);
void rtc_WriteSec(void);
void rtc_WriteHrs(void);
void rtc_WriteMin(void);
void rtc_WriteDom(void);
void rtc_WriteMon(void);
void rtc_WriteYar(void);
void rtc_ReadData(void);

void nvd_ReadData(void);
void nvd_UpdateCksum(void);
void nvd_WritePZTblSel(void);
void nvd_WriteSinkToneCps(void);
void nvd_WriteClimbToneCps(void);
void nvd_WriteFLogOldestRec(void);

code char* szMon[] = {"JAN","FEB","MAR","APR","MAY","JUN","JUL","AUG","SEP","OCT","NOV","DEC"};

code char gnFontTbl[310] = {	
                     0x00,0x00,0x00,0x00,0x00,	// 20 space		
							0x00,0x00,0x5f,0x00,0x00,	// 21 !			
							0x00,0x07,0x00,0x07,0x00,	// 22 "			
							0x14,0x7f,0x14,0x7f,0x14,	// 23 #			
							0x39,0x44,0x42,0x42,0x26,	// 24 $	Celsius		
							0x23,0x13,0x08,0x64,0x62,	// 25 %
							0x36,0x49,0x55,0x22,0x50,	// 26 &
							0x00,0x05,0x03,0x00,0x00,	// 27 '
							0x00,0x1c,0x22,0x41,0x00,	// 28 (
							0x00,0x41,0x22,0x1c,0x00,	// 29 )
							0x14,0x08,0x3e,0x08,0x14,	// 2a *
							0x08,0x08,0x3e,0x08,0x08,	// 2b +
							0x07,0x81,0xb7,0xD1,0x17,	// 2c ,   m/s
							0x08,0x08,0x08,0x08,0x08,	// 2d -
							0x00,0x60,0x60,0x00,0x00,	// 2e .
							0x20,0x10,0x08,0x04,0x02,	// 2f /
							0x3e,0x51,0x49,0x45,0x3e,	// 30 0
							0x00,0x42,0x7f,0x40,0x00,	// 31 1
							0x42,0x61,0x51,0x49,0x46,	// 32 2
							0x21,0x41,0x45,0x4b,0x31,	// 33 3
							0x18,0x14,0x12,0x7f,0x10,	// 34 4
							0x27,0x45,0x45,0x45,0x39,	// 35 5
							0x3c,0x4a,0x49,0x49,0x30,	// 36 6
							0x01,0x71,0x09,0x05,0x03,	// 37 7
							0x36,0x49,0x49,0x49,0x36,	// 38 8
							0x06,0x49,0x49,0x29,0x1e,	// 39 9
							0x00,0x36,0x36,0x00,0x00,	// 3a :
							0x00,0x56,0x36,0x00,0x00,	// 3b ;
							0x08,0x14,0x22,0x41,0x00,	// 3c <
							0x14,0x14,0x14,0x14,0x14,	// 3d =
							0x00,0x41,0x22,0x14,0x08,	// 3e >
							0x02,0x01,0x51,0x09,0x06,	// 3f ?
							0x32,0x49,0x79,0x41,0x3e,	// 40 @
							0x20,0x54,0x54,0x54,0x78,	// 41 a
							0x7f,0x48,0x44,0x44,0x38,	// 42 b
							0x38,0x44,0x44,0x44,0x20,	// 43 c
							0x38,0x44,0x44,0x48,0x7f,	// 44 d
							0x38,0x54,0x54,0x54,0x18,	// 45 e
							0x08,0x7e,0x09,0x01,0x02,	// 46 f
							0x0c,0x52,0x52,0x52,0x3e,	// 47 g
							0x7f,0x08,0x04,0x04,0x78,	// 48 h
							0x00,0x44,0x7d,0x40,0x00,	// 49 i
							0x20,0x40,0x44,0x3d,0x00,	// 4a j 
							0x7f,0x10,0x28,0x44,0x00,	// 4b k
							0x00,0x41,0x7f,0x40,0x00,	// 4c l
							0x7c,0x04,0x18,0x04,0x78,	// 4d m
							0x7c,0x08,0x04,0x04,0x78,	// 4e n
							0x38,0x44,0x44,0x44,0x38,	// 4f o
							0x7c,0x14,0x14,0x14,0x08,	// 50 p
							0x08,0x14,0x14,0x18,0x7c,	// 51 q
							0x7c,0x08,0x04,0x04,0x08,	// 52 r
							0x48,0x54,0x54,0x54,0x20,	// 53 s
							0x04,0x3f,0x44,0x40,0x20,	// 54 t
							0x3c,0x40,0x40,0x20,0x7c,	// 55 u
							0x1c,0x20,0x40,0x20,0x1c,	// 56 v
							0x3c,0x40,0x30,0x40,0x3c,	// 57 w
							0x44,0x28,0x10,0x28,0x44,	// 58 x
							0x0c,0x50,0x50,0x50,0x3c,	// 59 y
							0x44,0x64,0x54,0x4c,0x44,	// 5a z
							0x7f,0x7f,0x3e,0x1c,0x08,	// 5b [   selected
							0x02,0x04,0x08,0x10,0x20,	// 5c Yen 
							0x7f,0x81,0x81,0x81,0x7f};	// 5d ]   

code s32 gnPZTbl[3][PZLUT_ENTRIES] = 
{{// 30  : Po = 100325 
-21098,
-12626,
-4086,
4525,
13208,
21964,
30795,
39702,
48687,
57750,
66894,
76121,
85431,
94827,
104310,
113883,
123547,
133303,
143155,
153104,
163153,
173303,
183558,
193918,
204388,
214969,
225664,
236476,
247408,
258463,
269644,
280954,
292397,
303976,
315694,
327557,
339567,
351729,
364047,
376526,
389171,
401986,
414977,
428149,
441508,
455060,
468811,
482769,
496939,
511330,
525949,
540805,
555906,
571262,
586882,
602777
},
{// 50 : Po = 101325
-12700,
-4244,
4279,
12874,
21541,
30281,
39095,
47986,
56953,
66000,
75126,
84335,
93628,
103006,
112472,
122026,
131672,
141410,
151244,
161174,
171204,
181335,
191570,
201911,
212361,
222922,
233597,
244388,
255300,
266334,
277494,
288782,
300204,
311761,
323457,
335297,
347285,
359424,
371719,
384174,
396795,
409586,
422552,
435700,
449033,
462560,
476285,
490216,
504360,
518724,
533316,
548144,
563216,
578543,
594134,
609999
},
{// 70  : Po = 102325
-4400,
4039,
12548,
21127,
29777,
38501,
47299,
56172,
65123,
74153,
83262,
92454,
101730,
111090,
120538,
130075,
139702,
149422,
159237,
169149,
179160,
189272,
199488,
209810,
220240,
230782,
241437,
252208,
263099,
274113,
285252,
296519,
307919,
319455,
331130,
342948,
354913,
367029,
379301,
391733,
404331,
417098,
430040,
443162,
456471,
469972,
483672,
497577,
511695,
526032,
540596,
555397,
570441,
585739,
601301,
617136
}};


#ifdef PROFILE
s32 tck1,tck2;
u8 cnt;
#endif

void main(void)
   {
   mcu_Init();
   rtc_Init();
	lcd_Init();
   nvd_ReadData();
   sns_Init();

#ifdef PROFILE
   tck1 = gnTimeTick;
   cnt = 255;
   while (cnt--) {
      sns_ConvertPressureTemperature(0);
      sns_ConvertPressureTemperature(1);
      gZBuf[gnSmpCnt] = var_Pa2Cm((gP[0]+gP[1])*5);
      gnTempCx10 = (gT[0]+gT[1])/2;
      var_AverageAltitude();
      var_ClimbRate();
#ifdef VARIO_DAMPING
      var_LPFClimbRate();
#endif
      }

   tck1 = gnTimeTick - tck1;
   tck1 = (tck1*10)/255;

   tck2 = gnTimeTick;
   cnt = 255;
   while (cnt--) {
     gnDispCnt++;
     switch(gnDispCnt&3)
        {
        case 0 : 
        default : ui_PrintAltitude(gszBuf,gnAvgAltM); lcd_SetPos(0,0); lcd_PrintSzX2(gszBuf);
                  ui_PrintClimbRate(gszBuf,gnCps); lcd_SetPos(3,48); lcd_PrintSz(gszBuf); break;
        case 1 :  ui_PrintTemperature(gszBuf,gnTempCx10); lcd_SetPos(1,64); lcd_PrintSz(gszBuf); 
                  ui_PrintRealTime(gszBuf); lcd_SetPos(5,0); lcd_PrintSz(gszBuf); break;
        case 2 :  if (gbFlightLog) {sprintf(gszBuf,"%bu:%02bu",gnFlightHrs,gnFlightMin);lcd_SetPos(5,60);lcd_PrintSz(gszBuf);
                  ui_PrintAltitudeAL(gszBuf,gnAvgAltM,gnLaunchAltM);lcd_SetPos(3,0);lcd_PrintSz(gszBuf); } break;
        }
     }
  tck2 = gnTimeTick - tck2;
  tck2 = (tck2*10)/255;

  lcd_ClearScreen();
  sprintf(gszBuf,"T1:%ld T2:%ld", tck1,tck2);
  lcd_PrintSz(gszBuf);
  while(1);
#endif

   ui_Sleep();

	while(1)
	   {
      switch (gnUIMode)
         {
         case UIMODE_FLOG :
         default :
         ui_FLogEventHandler();
         break;

         case UIMODE_SET :
         ui_SetEventHandler();
         break;

         case UIMODE_RUN :
         ui_RunEventHandler();
         break;

         case UIMODE_SLEEP :
         ui_SleepHandler();
         break;
         }
	   }
   }



void nvd_ReadData(void)
   {
   u8 buf[5],cksum;
#ifdef FLOG_USE_SERIAL_EEPROM
   u8 cnt;
#endif
   smb_rbuf(SMB_ID_DS1338,buf,8,5);
   cksum = buf[0]+buf[1]+buf[2]+buf[3];
   if (cksum == buf[4])
      {
      gnPZTblSel = buf[NVD_ADDR_PZTBL_SEL];
      gnSinkToneCps = 10*(s16)buf[NVD_ADDR_SINKTONE_CPS];
      gnClimbToneCps = (s16)buf[NVD_ADDR_CLIMBTONE_CPS];
      gnFLogOldestRec = buf[NVD_ADDR_FLOG_OLDEST_REC];
      }
   else
      {
      lcd_SetPos(0,0);
      lcd_PrintSz("MEMORY CORRUPT");
      lcd_SetPos(1,0);
      lcd_PrintSz("(SEL)  CLEAR");
      while(!gbBtnSPressed);
      gbBtnSPressed = 0;
      lcd_SetPos(2,0);
      lcd_PrintSz("CLEARING...");
      gnPZTblSel = 1; 
      nvd_WritePZTblSel();
      gnSinkToneCps = SINKTONE_CPS_DEF;
      nvd_WriteSinkToneCps();
      gnClimbToneCps = CLIMBTONE_CPS_DEF;
      nvd_WriteClimbToneCps();
      gnFLogOldestRec = 0;
      nvd_WriteFLogOldestRec();
#ifdef FLOG_USE_FLASH
      //flash_ErasePage(FLOG_PAGE_ADDR);
      //flash_Fill(FLOG_PAGE_ADDR,FLASH_PAGESIZE,0);
#endif
#ifdef FLOG_USE_SERIAL_EEPROM
      cnt = FLOG_REC_BYTES*FLOG_MAX_FLIGHTS;
      while (cnt--) smb_wbyte(SMB_ID_24LC02,FLOG_EEP_ADDR+cnt,0);
#endif
      lcd_SetPos(3,0);
      lcd_PrintSz("DONE!");
      tmr_WaitTicks(100);
      }
   }

void nvd_UpdateCksum(void)
   {
   u8 buf[4],cksum;
   smb_rbuf(SMB_ID_DS1338,buf,8,4);
   cksum = buf[0]+buf[1]+buf[2]+buf[3];
   smb_wbyte(SMB_ID_DS1338,NVD_ADDR_CKSUM,cksum);
   }

void nvd_WritePZTblSel(void)
   {
   smb_wbyte(SMB_ID_DS1338,NVD_ADDR_PZTBL_SEL,gnPZTblSel);
   nvd_UpdateCksum();
   }

void nvd_WriteSinkToneCps(void)
   {
   smb_wbyte(SMB_ID_DS1338,NVD_ADDR_SINKTONE_CPS,(u8)(gnSinkToneCps/10));
   nvd_UpdateCksum();
   }

void nvd_WriteClimbToneCps(void)
   {
   smb_wbyte(SMB_ID_DS1338,NVD_ADDR_CLIMBTONE_CPS,(u8)gnClimbToneCps);
   nvd_UpdateCksum();
   }

void nvd_WriteFLogOldestRec(void)
   {
   smb_wbyte(SMB_ID_DS1338,NVD_ADDR_FLOG_OLDEST_REC,gnFLogOldestRec);
   nvd_UpdateCksum();
   }


void Tmr2_ISR(void) interrupt INTERRUPT_TIMER2
   {
   TMR2CN &= ~0x80;
   gbTmr2Flag = 1;
   gBtnLState = (gBtnLState<<1) | (u8)gpioBtnL | 0xF0;
   if (gBtnLState == 0xF8)
      {
      gbBtnLPressed = 1;
      }
   gBtnRState = (gBtnRState<<1) | (u8)gpioBtnR | 0xF0;
   if (gBtnRState == 0xF8)
      {
      gbBtnRPressed = 1;
      }
   gBtnSState = (gBtnSState<<1) | (u8)gpioBtnS | 0xF0;
   if (gBtnSState == 0xF8)
      {
      gbBtnSPressed = 1;
      }
   gBtnPState = (gBtnPState<<1) | (u8)gpioBtnP | 0xF0;
   if (gBtnPState == 0xF8)
      {
      gbBtnPPressed = 1;
      }
   if (gnUIMode != UIMODE_RUN) return;
   
   if (gnBeepPeriodTicks <= 0 || (ABS(gnCps-gnVarioCps) > BEEP_DISCR_THRESHOLD) || (gnCps > 0 && gnVarioCps <= 0) || (gnCps <= 0 && gnVarioCps > 0))
   	  {
   	  gnVarioCps = gnCps;
      if (gnVarioCps < -gnSinkToneCps)
         {
         PCA0CPH0  = 60;
         PCA0CPM0  |= 0x6;  // switch on tone by setting PCA0 to square wave frequency mode
         gnBeepPeriodTicks = 0;
         gnBeepTransition  = 0;
         }
      else
         {    
         if (gbStable)
            {        
            gnBeepCps  = (gbFlightLog ? gnVarioCps + gnClimbToneCps : 0);
            }
         else
            {
            gnBeepCps = gnVarioCps;
            }
         if (gnBeepCps > 0)  	   
            {
            gnBeepPeriodTicks =  (gnBeepCps > 624) ? 2 :  28 - (gnBeepCps/24);
            gnBeepFreqReg = (gnBeepCps > 850) ? 40 : 210 - (gnBeepCps/5);
            gnBeepTransition = (gnBeepPeriodTicks>>1);   
			if ((gnBeepPeriodTicks - gnBeepTransition) > 5) 
				gnBeepTransition = gnBeepPeriodTicks - 5;
            PCA0CPH0  = gnBeepFreqReg; // square wave frequency = 187500/gnBeepFreqReg
            PCA0CPM0  |= 0x6;  // switch on tone
            }
         else
            {
            PCA0CPM0 &= ~0x44; // switch off tone AND force CEX0 pin to 0 by setting PCA0 to 8bit pwm mode with 0% duty cycle (ECOM = 0)
            }
         }
   	}
   else
      {
      gnBeepPeriodTicks--;
      if (gnBeepPeriodTicks <= gnBeepTransition)
         {
         PCA0CPM0 &= ~0x44; // switch off
         }
      }       	
   }


void Tmr3_ISR(void) interrupt INTERRUPT_TIMER3
   {
   TMR3CN &= ~0x80;
   gnTimeTick++;
   gnSecTick++;
   if (gnSecTick == 100)
      {
      gnSecTick = 0;
      gnSec++;
      if (gnSec == 60)
         {
         gnSec = 0;
         gnMin++;
         if (gnMin == 60)
            {
            gnMin = 0;
            gnHrs++;
            if (gnHrs == 24)
               gnHrs = 0;
            }
         }
      
      if (gbFlightLog)
         {
         gnFlightSec++;
         if (gnFlightSec == 60)
            {
            gnFlightSec = 0;
            gnFlightMin++;
            if (gnFlightMin == 60)
               {
               gnFlightMin = 0;
               gnFlightHrs++;
               }
            }
         }
      }
   }



void sns_ReadSample(void)
   {
   sns_Reset();
   sns_TriggerTemperatureSample();
   while (gpioMDOUT1 || gpioMDOUT2);
   sns_ReadWord(&gwD2[0],&gwD2[1]); // read temperature
   sns_Reset();
   sns_TriggerPressureSample();
   while (gpioMDOUT1 || gpioMDOUT2);
   sns_ReadWord(&gwD1[0],&gwD1[1]); // read pressure
   sns_ConvertPressureTemperature(0);
   sns_ConvertPressureTemperature(1);
   gnAltCm = var_Pa2Cm((gP[0]+gP[1])*5);
   gnTempCx10 = (gT[0]+gT[1])/2;
   }

void sns_InitZBuf(void)
   {
   u8 cnt = SNS_NUM_Z_SAMPLES;
   sns_ReadSample();
   while (cnt--)
      {
      gZBuf[cnt] = gnAltCm;
      }
   var_AverageAltitude();
   }

void sns_Init(void)
   {
   gnSmpCnt = 0;
   gnAvgAltM = 0;
   gnCps = gnCpsx32 = 0;
   gpioMDOUT1 = 1; // input pin, set high impedance
   gpioMDOUT2 = 1; // input pin, set high impedance
   sns_Reset();
   sns_ReadCoeffs();
   sns_InitZBuf();
   }


// 1010101010101010 + 00000
void sns_Reset(void)
   {
   sns_SendCmd(0x155540,21);
   }


void sns_ReadWord(u16* pw1, u16* pw2)
   {
   u8 clk = 16;
   *pw1 = 0;
   *pw2 = 0;
   while (clk--)
      {
      gpioMSCLK = 1; 
      gpioMSCLK = 0;
      *pw1 |=  (((u16)gpioMDOUT1) << clk);
      *pw2 |=  (((u16)gpioMDOUT2) << clk);
      }
   gpioMSCLK = 1; 
   gpioMSCLK = 0;
   return;
   }

// ms bit first
void sns_SendCmd(u32 cmd, u8 nbits)
   {
   while (nbits--)
      {
      gpioMDIN = (cmd & (u32)(1<<nbits))  ? 1 : 0;
      gpioMSCLK = 1; 
      gpioMSCLK = 0;
      }
   }

// 111 + 6bit coeff addr + 000 + 1clk(send0) 
void sns_ReadCoeff(u8 addr, u16* pc1, u16* pc2)
   {
   u32 cmd = (u32)0x1C00 | (((u32)addr)<<4);
   sns_SendCmd(cmd,13);
   sns_ReadWord(pc1, pc2);
   }


void sns_ReadCoeffs(void)
   {
   u16 wa,wb,wc,wd;
   sns_ReadCoeff(0x15,&wa,&wb);
   sns_ReadCoeff(0x16,&wc,&wd);
   gwC1[0] = (u16)((wa >> 1) & (u16)0x7FFF);
   gwC5[0] = (u16)(((wa & 0x1)<<10) | ((wc >> 6) & (u16)0x3FF));
   gwC6[0] = (u16)(wc & 0x3F);

   gwC1[1] = (u16)((wb >> 1) & (u16)0x7FFF);
   gwC5[1] = (u16)(((wb & 0x1)<<10) | ((wd >> 6) & (u16)0x3FF));
   gwC6[1] = (u16)(wd & 0x3F);

   sns_ReadCoeff(0x19,&wa,&wb);
   sns_ReadCoeff(0x1A,&wc,&wd);
   gwC4[0] = (u16)((wa>>6) & 0x3FF);
   gwC2[0] = (u16)(((wa & 0x3F) << 6) | (wc & 0x3F));
   gwC3[0] =  (u16)((wc >> 6) & 0x3FF);

   gwC4[1] = (u16)((wb>>6) & 0x3FF);
   gwC2[1] = (u16)(((wb & 0x3F) << 6) | (wd & 0x3F));
   gwC3[1] =  (u16)((wd >> 6) & 0x3FF);
   //sprintf(gszBuf,"%u %u %u ",gwC1[0],gwC2[0],gwC3[0]);
   //sprintf(gszBuf,"%u %u %u",gwC4[0],gwC5[0],gwC6[0]);
   }


// 111 + 1010 + 000 + 2clks(send 0)
void sns_TriggerPressureSample(void)
   {
   sns_SendCmd(0xF40,12);
   }

// 111 + 1001 + 000 + 2clks(send 0)
void sns_TriggerTemperatureSample(void)
   {
   sns_SendCmd(0xF20,12);
   }


void sns_ConvertPressureTemperature(u8 sel)
    {
    s32 UT1,dT,OFF,SENS,X,TEMP,P2,T2;
    UT1  = (gwC5[sel]<<3) + 20224;
    dT   = (s32)gwD2[sel] - UT1;
    TEMP = 200 + ((dT*(gwC6[sel]+50))>>10);
    OFF  = (gwC2[sel]<<2) + (((gwC4[sel]-512)*dT)>>12);
    SENS = gwC1[sel] + ((gwC3[sel]*dT)>>10)  + 24576;
    X    = ((SENS* ((s32)gwD1[sel] - 7168))>>14)  - OFF;
    gP[sel] = ((X*10)>>5) + 2500;
    if (TEMP < 200)
        {
        T2 = (11*(gwC6[sel]+24)*(200-TEMP)*(200-TEMP))>>20;
        P2 = (3*T2*(gP[sel]-3500))>>14;
        }
    else
        {
        T2 = 0;
        P2 = 0;
        }
    TEMP = TEMP-T2;
    gP[sel] = gP[sel]-P2;
    gT[sel] = (s16)TEMP;  
    }

// MS5534B has 10pA resolution  = ~80cm altitude resolution at sea level
// Averaging the pressure readings of the two sensors gives us an additional 
// bit of pressure resolution => 40cm altitude resolution

void sns_ReadSampleContinuous(void)
   {  
   if (gpioMDOUT1 || gpioMDOUT2) return;  // conversion not yet complete
   if (gnAcqState)
      {
      sns_ReadWord(&gwD1[0],&gwD1[1]); // read pressure
      sns_Reset();
      sns_TriggerTemperatureSample();
      sns_ConvertPressureTemperature(0);
      sns_ConvertPressureTemperature(1);
      // ms5534 has 0.1mbar (10pa) resolution, LUT has 1 pa input resolution. 
      gZBuf[gnSmpCnt] = gnAltCm = var_Pa2Cm((gP[0]+gP[1])*5);
      gnTempCx10 = (gT[0]+gT[1])/2;
      var_AverageAltitude();
      var_ClimbRate();
#ifdef VARIO_DAMPING
      var_LPFClimbRate();
#endif
      gnSmpCnt++;
      if (gnSmpCnt == SNS_NUM_Z_SAMPLES)
         {
         gnSmpCnt = 0;
         }
      gnAcqState = 0;
      }
   else
      {
      gnDispCnt++;
      sns_ReadWord(&gwD2[0],&gwD2[1]); // read temperature
      sns_Reset();
      sns_TriggerPressureSample();
      if (gnUIMode == UIMODE_RUN)
         {
         switch(gnDispCnt&3)
            {
            case 0 : 
            default : ui_PrintAltitude(gszBuf,gnAvgAltM); lcd_SetPos(0,0); lcd_PrintSzX2(gszBuf);
                      ui_PrintClimbRate(gszBuf,gnCps); lcd_SetPos(3,48); lcd_PrintSz(gszBuf); break;
            case 1 :  ui_PrintTemperature(gszBuf,gnTempCx10); lcd_SetPos(1,64); lcd_PrintSz(gszBuf); 
                      ui_PrintRealTime(gszBuf); lcd_SetPos(5,0); lcd_PrintSz(gszBuf); break;
            case 2 :  if (gbFlightLog) {sprintf(gszBuf,"%02bu:%02bu",gnFlightHrs,gnFlightMin);if (gszBuf[0]=='0')gszBuf[0]=' ';lcd_SetPos(5,53);lcd_PrintSz(gszBuf);
                      ui_PrintAltitudeAL(gszBuf,gnAvgAltM,gnLaunchAltM);lcd_SetPos(3,0);lcd_PrintSz(gszBuf); }break;
            }
         }
      gnAcqState = 1;
      }
   }


void var_ClimbRate(void)
   {
   u8 n;
   s32 sum_zt,sum_z,sum_t,sum_t2,trel,num,den,tref,z;
   tref = gnSmpCnt; // time origin is the current sample in window
   sum_zt = sum_z = sum_t = sum_t2 = 0;
   n = SNS_NUM_Z_SAMPLES;
   while (n--)
      {
      z = gZBuf[n];
      trel = n-tref;
      if (n > tref) trel -= SNS_NUM_Z_SAMPLES;
      sum_z += z;
      sum_t += trel;
      sum_t2 += (trel*trel);
      sum_zt += (trel*z);
      }
   num = (SNS_NUM_Z_SAMPLES*sum_zt) - (sum_z*sum_t);
   den = (SNS_NUM_Z_SAMPLES*sum_t2) - (sum_t*sum_t);
   gnCps = den ? (s16)((num*SNS_ACQ_PER_SEC)/den) : 0;
   return;
   }


void var_AverageAltitude(void)
   {
   u8 cnt;
   s32 avg = 0;
   cnt = SNS_NUM_Z_SAMPLES;
   while (cnt--)
      {avg += gZBuf[cnt];}
   avg /= SNS_NUM_Z_SAMPLES;
   avg = ((avg > 0) ? (avg+50)/100 : (avg-50)/100); 
   gnAvgAltM = (s16)avg;
   return; 
   }
  
#ifdef VARIO_DAMPING  
void var_LPFClimbRate(void)
   {
   s32 tmp;
   tmp = gnCpsx32*(32L-KLPF) + (((s32)gnCps)<<5)*KLPF;
   gnCpsx32 = (tmp >= 0 ? ((tmp + 16L)>>5) : ((tmp - 16L)>>5));
   gnCps = (s16)(gnCpsx32>>5);
   }  
#endif

s32 var_Pa2Cm(s32 pa)
   {
   s32 inx,pa1,pa2,z1,z2,z;
   if (pa > PA_INIT)
      {
      z = gnPZTbl[gnPZTblSel][0];
      }
   else
      {
      inx = (PA_INIT - pa)>>10;      
      if (inx >= PZLUT_ENTRIES)
         {
         z = gnPZTbl[gnPZTblSel][PZLUT_ENTRIES-1];
         }
      else
         {
         pa2 = PA_INIT - (inx<<10);
         pa1 = (inx > 0 ? pa2+PA_DELTA : pa2);
         z1 = (inx > 0 ? gnPZTbl[gnPZTblSel][inx-1] : gnPZTbl[gnPZTblSel][0]);
         z2 = gnPZTbl[gnPZTblSel][inx];
         z = ((pa1-pa)*z2 + (pa-pa2)*z1)>>10;
         }
      }
   return z;
   }


void tmr_WaitTicks(u8 ticks)
   {
   s32 tick = gnTimeTick + ticks;
   while (gnTimeTick < tick);
   }

void ui_noLZU(char* szBuf)
   {
   u8 cnt;
   u8 nchars = strlen(szBuf);
   cnt = 0;
   while (cnt < nchars-1 && szBuf[cnt] == '0') szBuf[cnt++] = ' ';
   }

void ui_noLZS(char* szBuf)
   {
   u8 cnt;
   char sign;
   u8 nchars = strlen(szBuf);
   sign = szBuf[0];
   cnt = 0;
   while (cnt < nchars && (szBuf[cnt] == '0' || szBuf[cnt] == sign)) szBuf[cnt++] = ' ';
   szBuf[cnt-1] = sign;
   }

void ui_PrintAltitude(char* szBuf, s16 nAlt)
    {
    CLAMP(nAlt,MIN_ALTITUDE_M,MAX_ALTITUDE_M);
    if (nAlt == 0)
       sprintf(szBuf,"   0M");
    else
    if (nAlt < 0)
       {
       sprintf(szBuf,"-%03dM", ABS(nAlt));
       ui_noLZS(szBuf);
       }
    else
       {
       sprintf(szBuf,"%04dM", nAlt); 
       ui_noLZU(szBuf);
       }
    }

void ui_PrintAltitudeAL(char* szBuf, s16 nAlt, s16 nLaunchAlt)
   {
   s16 nAltAL;
   nAltAL = nAlt - nLaunchAlt;
   CLAMP(nAltAL,-9999,9999);
   if (nAltAL == 0)
       sprintf(szBuf,"    0M");
    else
       {
       sprintf(szBuf,"%s%04dM", nAltAL > 0 ? "+" : "-", ABS(nAltAL));
       ui_noLZS(szBuf);
       }
    }

void ui_PrintClimbRate(char* szBuf, s16 nCps)
   {
   s16 dec,frac;
   CLAMP(nCps, MIN_CLIMBRATE_CPS, MAX_CLIMBRATE_CPS);
   if (nCps >= 0) nCps += 5; else nCps -= 5;
   nCps /= 10;
   dec = ABS(nCps)/10;
   frac = ABS(nCps)%10;
   if (nCps == 0)
      {
      sprintf(szBuf,"  0.0,");
      }
   else
   if (nCps > 0)
      {
      sprintf(szBuf,"+%02d.%d,",dec, frac);
      }
   else
      {
      sprintf(szBuf,"-%02d.%d,",dec, frac);
      }
   if (szBuf[1]=='0') 
      {
      szBuf[1] = szBuf[0];
      szBuf[0] = ' ';
      }
   }
       
void ui_PrintTemperature(char* szBuf, s16 nTemp)
   {
   if (nTemp >= 0) 
      nTemp += 5;
   else 
      nTemp -= 5;
   nTemp /= 10;
   CLAMP(nTemp,MIN_TEMP_C,MAX_TEMP_C);
   gnTempC = nTemp;
   if (nTemp == 0)
      {
      sprintf(szBuf," 0$");
      }
   else
   if (nTemp < 0)
      {
      sprintf(szBuf,"-%d$", ABS(nTemp));
      }
   else
      {
      sprintf(szBuf,"%02d$", nTemp);
      ui_noLZU(szBuf);
      }
   }


void ui_PrintRealTime(char* szBuf)
   {
   CLAMP(gnHrs,0,23);
   CLAMP(gnMin,0,59);

   if (gnHrs/12)
      {
      sprintf(szBuf,"%02bu:%02buPM",(gnHrs-12) ?  (gnHrs-12) : 12,gnMin);
      }
   else
   if (gnHrs < 10)
      {
      sprintf(szBuf," %bu:%02buAM",gnHrs,gnMin);
      }
   else
      {
      sprintf(szBuf,"%bu:%02buAM",gnHrs,gnMin);
      }
   if (szBuf[0] == '0') szBuf[0] = ' ';
   }


void var_updateFlightData(void)
   {
   if (gnAvgAltM > gnMaxAltM) gnMaxAltM = gnAvgAltM;
   if (gnCps > gnMaxCps) gnMaxCps = gnCps;
   if (-gnCps > gnMinCps) gnMinCps = -gnCps;
   if (gnTempC > gnMaxTempC) gnMaxTempC = gnTempC;
   if (gnTempC < gnMinTempC) gnMinTempC = gnTempC;
   }



void mcu_Init(void)
   {
    PCA0CN    = 0x40;
    PCA0MD    &= ~0x40;
    PCA0MD    = 0x02;
    PCA0CPM0  = 0x46;
    PCA0CPH0  = 0xFA;
    PCA0CPM0  &= ~0x44;

    TCON      = 0x40;
    TMOD      = 0x22;
    CKCON     = 0x55;
    TH0       = 0xCE;
    TH1       = 0x64;
    TMR2CN    = 0x04;
    TMR2RLL   = 0xEC;
    TMR2RLH   = 0x32;
    TMR2L     = 0xFF;
    TMR2H     = 0xFF;
    TMR3CN    = 0x04;
    TMR3RLL   = 0x68;
    TMR3RLH   = 0xC5;
    TMR3L     = 0xFF;
    TMR3H     = 0xFF;

    SMB0CF    = 0xD1;

    SPI0CFG   = 0x40;

    // P0.0  -  Skipped,     Open-Drain, Digital
    // P0.1  -  Skipped,     Open-Drain, Digital
    // P0.2  -  Skipped,     Push-Pull,  Digital
    // P0.3  -  Skipped,     Open-Drain, Digital
    // P0.4  -  TX0 (UART0), Push-Pull,  Digital
    // P0.5  -  RX0 (UART0), Open-Drain, Digital
    // P0.6  -  Skipped,     Push-Pull,  Digital
    // P0.7  -  Skipped,     Push-Pull,  Digital

    // P1.0  -  Skipped,     Push-Pull,  Digital
    // P1.1  -  Skipped,     Open-Drain, Digital
    // P1.2  -  Skipped,     Open-Drain, Digital
    // P1.3  -  SDA (SMBus), Open-Drain, Digital
    // P1.4  -  SCL (SMBus), Open-Drain, Digital
    // P1.5  -  CEX0 (PCA),  Push-Pull,  Digital
    // P1.6  -  Skipped,     Open-Drain, Digital
    // P1.7  -  Skipped,     Open-Drain, Digital
    // P2.0  -  Skipped,     Open-Drain, Digital
    // P2.1  -  Skipped,     Open-Drain, Digital
    // P2.2  -  Skipped,     Push-Pull,  Digital
    // P2.3  -  Skipped,     Push-Pull,  Digital

    // P2.6     Analog input

    P2MDIN    = 0xBF;
    P0MDOUT   = 0xD4;
    P1MDOUT   = 0x21;
    P2MDOUT   = 0x0C;
    P0SKIP    = 0xCF;
    P1SKIP    = 0xC7;
    P2SKIP    = 0x0F;
    XBR0      = 0x05;
    XBR1      = 0x41;

    // btn pins and ext clk pin are inputs, set high impedance
    gpioBtnL = 1; 
    gpioBtnR = 1;
    gpioBtnP = 1;
    gpioBtnS = 1;
    gpioExtClk = 1;

    AMX0P = 0x0E;   // ADC0 positive input = P2.6
    AMX0N = 0x1F;   // ADC0 negative input = GND
    REF0CN = 0x8;   // Vdd is VREF
                                       
    EIE1  = 0x81;
    IE    = 0xA0;  // enable all interrupts (EA), timer2 (ET2), 
   }



void lcd_Init(void)
   {
	gpioRST_ = 0;		
   tmr_WaitTicks(5);
	gpioRST_ = 1;		
	lcd_WriteCmd(0x21);		// Activate Chip and H=1.
	lcd_WriteCmd(0x87);		// Set LCD Voltage to about 7V.
	lcd_WriteCmd(0x13);		// Adjust voltage bias.
	lcd_WriteCmd(0x20);		// Horizontal addressing and H=0.
	lcd_ClearScreen();		// Erase all pixel on the DDRAM.
	lcd_WriteCmd(0x0C);		// Display Normal.
   lcd_ClearScreen();

   lcd_SetPos(0,0);
   lcd_PrintSz("ALTI-VARIO");
   lcd_SetPos(2,0);
   lcd_PrintSz("VERSION 1.01");
   bat_Voltage();
   lcd_SetPos(4,0);
   sprintf(gszBuf,"BATTERY %bu.%buV",gnVBatx10/10,gnVBatx10%10);
   lcd_PrintSz(gszBuf);
   tmr_WaitTicks(200);
   }


void lcd_WriteCmd(char command_in)
   {
	gpioDC = 0;				
	lcd_ClockData(command_in);	
   }


void lcd_WriteDat(char data_in)
   {
	gpioDC = 1;				
	lcd_ClockData(data_in);		
   }


void lcd_ClockData(char bits_in)
   {
	u8 bitcnt = 8;
   while (bitcnt--)
		{
		gpioSCLK = 0;								
		gpioSDA = ((bits_in&0x80) ? 1 : 0);		// MSb first
		gpioSCLK = 1;								// Data is clocked on the rising edge
		bits_in=bits_in<<1;						
		}
   }


void lcd_ClearScreen(void)
   {
	s16 cnt = 504;  //84 cols x 6 byte rows
	lcd_SetPos(0,0);										
   while (cnt--) lcd_WriteDat(0);
   }

	
void lcd_SetPos(char y, char x)
   {  
	lcd_WriteCmd(0x80|(x&0x7f));	// X axis
	lcd_WriteCmd(0x40|(y&0x07));	// Y axis
   gnLCDX = x;
   gnLCDY = y;
   }


void lcd_PutChar(char ch)
   {
	s16 inx;
   u8 cnt, col;
	if ((ch < 0x20) || (ch > 0x5d)) return;
	inx = 5*(ch - 32);
   cnt = 5;
	while(cnt--)
	   {
		col = gnFontTbl[inx];
		lcd_WriteDat(col);
		inx++;
	   }
   // 1 pixel horizontal spacing
	lcd_WriteDat(0x00); 
   }


void lcd_PutCharX2 (char ch )
   {
   s16 inx;
   u8 cnt,col,b1,b2;
	if ((ch < 0x20) || (ch > 0x5d)) return;	
	inx = 5*(ch - 32);
   cnt = 5;
   while (cnt--)
      {
      col = gnFontTbl[inx];
      b1 =  (col & 0x01) * 3;
      b1 |= (col & 0x02) * 6;
      b1 |= (col & 0x04) * 12;
      b1 |= (col & 0x08) * 24;
      col >>= 4;
      b2 =  (col & 0x01) * 3;
      b2 |= (col & 0x02) * 6;
      b2 |= (col & 0x04) * 12;
      b2 |= (col & 0x08) * 24;
		lcd_WriteDat(b1);
		lcd_WriteDat(b1);
      lcd_SetPos(gnLCDY+1,gnLCDX);
		lcd_WriteDat(b2);
		lcd_WriteDat(b2);
      lcd_SetPos(gnLCDY-1,gnLCDX+2);
      inx++;
      }
   // 1 pixel horizontal spacing
   lcd_WriteDat(0);
   lcd_SetPos(gnLCDY+1,gnLCDX);
   lcd_WriteDat(0);
   lcd_SetPos(gnLCDY-1,gnLCDX+1);
   }


void lcd_PrintSz(char *sz)
   {
	while(*sz) lcd_PutChar(*sz++);								
   }


void lcd_PrintSzX2(char *sz)
   {
	while(*sz) lcd_PutCharX2(*sz++);								
   }


// SMBus ISR state machine
// - Master only implementation - no slave or arbitration states defined
// - All incoming data is written starting at the global pointer <gpSmbDataRcv>
// - All outgoing data is read from the global pointer <gpSmbDataXmt>

void smb_ISR (void) interrupt INTERRUPT_SMBUS0
   {
   static u8 i;             // Used by the ISR to count the number of data bytes sent or received
   bit bFail = 0;                       // Used by the ISR to flag failed transfers
   static bit bSendStart = 0;          // Send a start
   switch (SMB0CN & 0xF0)              // Status vector
      {
      // Master Transmitter/Receiver: sns_start condition transmitted.
      case SMB_MTSTA:
         SMB0DAT = gnSmbDevId;             // Load address of the target slave
         SMB0DAT &= 0xFE;              // Clear the LSB of the address for the R/W bit
         SMB0DAT |= gbSmbRW;            // Load R/W bit
         STA = 0;                      // Manually clear sns_start bit
         i = 0;                        // Reset data byte counter
         break;
      // Master Transmitter: Data byte (or Slave Address) transmitted
      case SMB_MTDB:
         if (ACK)                      // Slave Address or Data Byte
            {                             // Acknowledged?
            if (bSendStart)
               {
               STA = 1;
               bSendStart = 0;
               break;
               }
            if(gbSmbSendAddr)       // Are we sending the word address?
               {
               gbSmbSendAddr = 0;   // Clear flag
               SMB0DAT = gnSmbAddr;    // Send word address
               if (gbSmbRandomRead)
                  {
                  bSendStart = 1;      // Send a sns_start after the next ACK cycle
                  gbSmbRW = SMB_READ;
                  }
               break;
               }
            if (gbSmbRW==SMB_WRITE)         // Is this transfer a SMB_WRITE?
               {
               if (i < gnSmbDataLen)   // Is there data to send?
                  {
                  SMB0DAT = *gpSmbDataXmt;                  // send data byte
                  gpSmbDataXmt++;                  // increment data out pointer
                  i++;                  // increment number of bytes sent
                  }
               else
                  {
                  STO = 1;              // Set STO to terminte transfer
                  gbSmbBusy = 0;         // Clear software busy flag
                  }
               }
            else {}                    // If this transfer is a READ, then take no action. Slaveaddress was transmitted. A separate 'case' is defined
                                       // for data byte recieved.
            }
         else                          // If slave NACK,
            {
            if(gbSmbAckPoll)
               {
               STA = 1;                // Restart transfer
               }
            else
               {
               bFail = 1;               // Indicate failed transfer
               }                          // and handle at end of ISR
            }
         break;
      // Master Receiver: byte received
      case SMB_MRDB:
         if ( i < gnSmbDataLen )       // Is there any data remaining?
            {
            *gpSmbDataRcv = SMB0DAT;   // Store received byte
            gpSmbDataRcv++;            // Increment data in pointer
            i++;                       // Increment number of bytes received
            ACK = 1;                   // Set ACK bit (may be cleared later in the code)
            }
         if (i == gnSmbDataLen)        // This is the last byte
            {
            gbSmbBusy = 0;              // Free SMBus interface
            ACK = 0;                   // Send NACK to indicate last byte of this transfer
            STO = 1;                   // Send sns_stop to terminate transfer
            }
         break;
      default:
         bFail = 1;                     // Indicate failed transfer and handle at end of ISR
         break;
      }
   if (bFail)                           // If the transfer failed,
      {
      SMB0CF &= ~0x80;                 // Reset communication
      SMB0CF |= 0x80;
      STA = 0;
      STO = 0;
      ACK = 0;
      gbSmbBusy = 0;                    // Free SMBus
      bFail = 0;
      }
   SI = 0;                             // Clear interrupt flag
   }

   

void smb_wbyte(u8 devId, u8 addr, u8 dat)
   {
   while (gbSmbBusy);                  
   gbSmbBusy = 1;                      // Claim SMBus
   gnSmbDevId = devId;                 // Set target slave address
   gbSmbRW = SMB_WRITE;                // Mark next transfer as a write
   gbSmbSendAddr = 1;                  // Send Word Address after Slave Address
   gbSmbRandomRead = 0;                // Do not send a START signal after the word address
   gbSmbAckPoll = 1;                   // Enable Acknowledge Polling (The ISR will automatically restart the transfer if the slave does not acknoledge its address.
   gnSmbAddr = addr;                   // Set the target address in the device internal memory space
   gnSmbData = dat;                    // Store <dat> (local variable) in a global variable so the ISR can read it after this function exits
   gpSmbDataXmt = &gnSmbData;          // The outgoing data pointer points to the <dat> variable
   gnSmbDataLen = 1;                   // Specify to ISR that the next transfer will contain one data byte
   STA = 1;                            // Initiate SMBus Transfer
   }
   

u8 smb_rbyte(u8 devId, u8 addr)
   {
   u8 retval;        
   while (gbSmbBusy);            
   gbSmbBusy = 1;                      // Claim SMBus
   gnSmbDevId = devId;                 // Set target slave address
   gbSmbRW = SMB_WRITE;                // A random read starts as a write then changes to a read after the repeated start is sent. The ISR handles this switchover if the <gbSmbRandomRead> bit is set.
   gbSmbSendAddr = 1;                  // Send Word Address after Slave Address
   gbSmbRandomRead = 1;                // Send a START after the word address
   gbSmbAckPoll = 1;                   // Enable Acknowledge Polling
   gnSmbAddr = addr;                   // Set the target address in the device internal memory space
   gpSmbDataRcv = &retval;             // The incoming data pointer points to the <retval> variable.
   gnSmbDataLen = 1;                   // Specify to ISR that the next transfer will contain one data byte
   STA = 1;                            // Initiate SMBus Transfer
   while(gbSmbBusy);                   // Wait until data is read
   return retval;
   }


void smb_rbuf(u8 devId, u8* pBuf, u8 addr, u8 len)
   {
   while (gbSmbBusy);                   
   gbSmbBusy = 1;                      // Claim SMBus
   gnSmbDevId = devId;                 // Set target slave address
   gbSmbRW = SMB_WRITE;                // A random read starts as a write then changes to a read after the repeated start is sent. The ISR handles this switchover if the <gbSmbRandomRead> bit is set.
   gbSmbSendAddr = 1;                  // Send Word Address after Slave Address
   gbSmbRandomRead = 1;                // Send a sns_start after the word address
   gbSmbAckPoll = 1;                   // Enable Acknowledge Polling
   gnSmbAddr = addr;                   // Set the target address in the devices internal memory space
   gpSmbDataRcv = (u8*) pBuf;          // Set the data pointer to receive buffer
   gnSmbDataLen = len;                 // Specify to ISR that the next transfer will contain <len> data bytes
   STA = 1;                            // Initiate SMBus Transfer
   while(gbSmbBusy);                   // Wait until data is read
   }

void rtc_Init(void)
   {
   xdata u8 b;
   b = smb_rbyte(SMB_ID_DS1338,0x0);
   // check CH bit is cleared to ensure oscillator is enabled
   if (b & 0x80)
      {
      smb_wbyte(SMB_ID_DS1338,0x0,0);
      }
   // enable 32.768kHz square wave output (sampling clock for MS5534B)
   smb_wbyte(SMB_ID_DS1338,0x7,0x13);
   }

void rtc_ReadData(void)
   {
   xdata u8 buf[7];
   smb_rbuf(SMB_ID_DS1338,buf,0x0,7);
   gnSec = (buf[0]&0xF) + 10*((buf[0]>>4)&7);
   gnMin = (buf[1]&0xF) + 10*((buf[1]>>4)&7);
   if (buf[2]&0x40)
      {
      gbHrsMode = RTC_HRS_12;
      gnHrs = (buf[2]&0xF) + 10*((buf[2]>>4)&1);
      CLAMP(gnHrs,0,12);
      gbAmPm = (buf[2]>>5)&1;
      }
   else
      {
      gbHrsMode = RTC_HRS_24;
      gnHrs = (buf[2]&0xF) + 10*((buf[2]>>4)&3);
      CLAMP(gnHrs,0,23);
      gbAmPm = gnHrs > 12 ? RTC_PM : RTC_AM;
      }
   gnDom = (buf[4]&0xF) + 10*((buf[4]>>4)&3);
   CLAMP(gnDom,1,31);
   gnMon = (buf[5]&0xF) + 10*((buf[5]>>4)&1);
   CLAMP(gnMon,1,12);
   gnYar = (buf[6]&0xF) + 10*(buf[6]>>4);
   CLAMP(gnYar,0,99);
   }

void rtc_WriteSec(void)
    {
    u8 b;
    b = (u8)((gnSec%10) + ((gnSec/10)<<4));
    smb_wbyte(SMB_ID_DS1338,RTC_ADDR_SEC,b);
    }

void rtc_WriteMin(void)
    {
    u8 b;
    b = (u8)((gnMin%10) + ((gnMin/10)<<4));
    smb_wbyte(SMB_ID_DS1338,RTC_ADDR_MIN,b);
    }

void rtc_WriteHrs(void)
    {
    u8 b;
    b = (u8)((gnHrs%10) + ((gnHrs/10)<<4) + (((u8)gbHrsMode)<<5));
    smb_wbyte(SMB_ID_DS1338,RTC_ADDR_HRS,b);
    }


void rtc_WriteDom(void)
    {
    u8 b;
    b = (u8)((gnDom%10) + ((gnDom/10)<<4));
    smb_wbyte(SMB_ID_DS1338,RTC_ADDR_DOM,b);
    }

void rtc_WriteMon(void)
    {
    u8 b;
    b = (u8)((gnMon%10) + ((gnMon/10)<<4));
    smb_wbyte(SMB_ID_DS1338,RTC_ADDR_MON,b);
    }

void rtc_WriteYar(void)
    {
    u8 b;
    b = (u8)((gnYar%10) + ((gnYar/10)<<4));
    smb_wbyte(SMB_ID_DS1338,RTC_ADDR_YAR,b);
    }



void ui_SetEventHandler(void)
   {
   if (gbTmr2Flag)
      {
      gbTmr2Flag = 0;
      ui_DisplaySettings();
      }
   if (gbBtnPPressed)
      {
      gbBtnPPressed = 0;
      ui_InitRunMode();
      }
   else
   if (gbBtnSPressed)
      {
      gbBtnSPressed = 0;
      gnParSel = (gnParSel < SEL_RTC_YAR ? gnParSel+1 : SEL_ALTITUDE);
      lcd_ClearScreen();
      ui_DisplaySettings();
      }
   else
   if (gbBtnLPressed)
      {
      gbBtnLPressed = 0;
      switch (gnParSel)
         {
         case SEL_ALTITUDE :  default : gnPZTblSel--; if (gnPZTblSel == 255) gnPZTblSel = 2; nvd_WritePZTblSel(); sns_ReadSample();break;
         case SEL_SINKTONE : gnSinkToneCps += 50; if (gnSinkToneCps > SINKTONE_CPS_MAX) gnSinkToneCps = SINKTONE_CPS_MAX; nvd_WriteSinkToneCps();break;
         case SEL_CLIMBTONE : gnClimbToneCps += 25; if (gnClimbToneCps > CLIMBTONE_CPS_MAX) gnClimbToneCps = CLIMBTONE_CPS_MAX; nvd_WriteClimbToneCps();break;
         case SEL_RTC_HRS : gnHrs--; if (gnHrs == 255) gnHrs = 23; rtc_WriteHrs(); break;
         case SEL_RTC_MIN : gnMin--; if (gnMin == 255) gnMin = 59; rtc_WriteMin(); gnSec = 0; rtc_WriteSec(); break;
         case SEL_RTC_DOM : gnDom--; if (gnDom < 1) gnDom = 31; rtc_WriteDom(); break;
         case SEL_RTC_MON : gnMon--; if (gnMon < 1) gnMon = 12; rtc_WriteMon(); break;
         case SEL_RTC_YAR : gnYar--; if (gnYar == 255) gnYar = 99; rtc_WriteYar(); break;
         }
      ui_DisplaySettings();
      }
   else
   if (gbBtnRPressed)
      {
      gbBtnRPressed = 0;
      switch (gnParSel)
         {
         case SEL_ALTITUDE :  default : gnPZTblSel++; if (gnPZTblSel > 2) gnPZTblSel = 0; nvd_WritePZTblSel(); sns_ReadSample();break;
         case SEL_SINKTONE : gnSinkToneCps -= 50; if (gnSinkToneCps < SINKTONE_CPS_MIN) gnSinkToneCps = SINKTONE_CPS_MIN; nvd_WriteSinkToneCps();break;
         case SEL_CLIMBTONE : gnClimbToneCps -= 25; if (gnClimbToneCps < CLIMBTONE_CPS_MIN) gnClimbToneCps = CLIMBTONE_CPS_MIN; nvd_WriteClimbToneCps();break;
         case SEL_RTC_HRS : gnHrs++; if (gnHrs > 23) gnHrs = 0; rtc_WriteHrs(); break;
         case SEL_RTC_MIN : gnMin++; if (gnMin > 59) gnMin = 0; rtc_WriteMin(); gnSec = 0; rtc_WriteSec(); break;
         case SEL_RTC_DOM : gnDom++; if (gnDom > 31) gnDom = 1; rtc_WriteDom(); break;
         case SEL_RTC_MON : gnMon++; if (gnMon > 12) gnMon = 1; rtc_WriteMon(); break;
         case SEL_RTC_YAR : gnYar++; if (gnYar > 99) gnYar = 0; rtc_WriteYar(); break;
         }
      ui_DisplaySettings();
      }
   }
   

void ui_FLogEventHandler(void)
   {
#ifdef FLOG_USE_SERIAL_EEPROM
   u8 cnt;
#endif
   if (gbBtnLPressed)
      {
      gbBtnLPressed = 0;
      gnFLogRec--;
      if (gnFLogRec == 255) gnFLogRec = FLOG_MAX_FLIGHTS-1;
      ui_DisplayFLog();
      }
   else
   if (gbBtnRPressed)
      {
      gbBtnRPressed = 0;
      gnFLogRec++;
      if (gnFLogRec >= FLOG_MAX_FLIGHTS) gnFLogRec = 0;
      ui_DisplayFLog();
      }
   else
   if (gbBtnSPressed)
      {
      gbBtnSPressed = 0;
      tmr_WaitTicks(20);
      if (gbBtnPPressed)
         {
         gbBtnPPressed = 0;
         lcd_ClearScreen();
         lcd_SetPos(4,0);
         lcd_PrintSz("ERASING LOG");
#ifdef FLOG_USE_FLASH
         flash_ErasePage(FLOG_PAGE_ADDR);
         flash_Fill(FLOG_PAGE_ADDR,FLASH_PAGESIZE,0);
#endif
#ifdef FLOG_USE_SERIAL_EEPROM
         cnt = FLOG_REC_BYTES*FLOG_MAX_FLIGHTS;
         while (cnt--) smb_wbyte(SMB_ID_24LC02,FLOG_EEP_ADDR+cnt,0);
#endif
         gnFLogOldestRec = 0;
         nvd_WriteFLogOldestRec();
         gnFLogRec = 0;
         lcd_ClearScreen();
         ui_DisplayFLog();
         }
      else
         {
         ui_InitSleepMode();
         }
      }
   else
   if (gbBtnPPressed)
      {
      gbBtnPPressed = 0;
      ui_InitSetMode();
      }
  }


void ui_RunEventHandler()
   {
   sns_ReadSampleContinuous();
/*
   if (gbBtnLPressed) {gbBtnLPressed = 0; gnCps -= 10;}
   else
   if (gbBtnRPressed) {gbBtnRPressed = 0; gnCps += 10;}
*/
   if (gbTmr2Flag) 
      {
      gbTmr2Flag = 0;
      if (gbFlightLog)
         {
         var_updateFlightData();
         }
      }     
   if (gbStable && !gbFlightLog && ((ABS(gnAvgAltM - gnLaunchAltM) > FLOG_THRESHOLD_M)))
      {
      gbFlightLog = 1; // flight start
      gnLaunchHrs = gnHrs;
      gnLaunchMin = gnMin;
      }
   if (gbBtnSPressed)
      {
      gbBtnSPressed = 0;
      if (!gbStable) gbStable = 1;
      if (!gbFlightLog) // register launch altitude, zero flight time
         {
         gnMaxAltM = gnLZAltM = gnLaunchAltM = gnAvgAltM;
         gnMaxCps = 0;
         gnMinCps = 0;
         gnMaxTempC = gnMinTempC = gnTempC;
         gnFlightHrs = 0;
         gnFlightMin = 0;
         lcd_SetPos(5,60);
         lcd_PrintSz(" OK ");
         }
      }
    else
    if (gbBtnPPressed)  
       {
       gbBtnPPressed = 0;
       PCA0CPM0  &= ~0x44;  // shut off vario tone
       PCA0CN    &= ~0x40;

       gbStable = 0;
       if (gbFlightLog) // if flight started, this marks end of flight
          {
          gbFlightLog = 0;
          gnLZAltM = gnAvgAltM;
          var_writeFlightLog();
          }
       ui_InitFLogMode();
       } 
   }


void ui_DisplaySettings(void)
    {
    u8 n;
    s8 offset;
    offset = gnParSel - 5;
    if (offset < 0) offset = 0;
    for (n = 0; n < 6; n++)
       {
       lcd_SetPos(n,0);
       if (gnParSel == (n+offset)) lcd_PrintSz("[");
       else lcd_PrintSz(" ");
       }
    n = 0;
    if (offset < 1)
       {sprintf(gszBuf, "ALT    %04d",(s16)((gnAltCm+50)/100));lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    if (offset < 2)
       {sprintf(gszBuf, "SINK  -%d.%d",gnSinkToneCps/100,(gnSinkToneCps%100)/10);lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    if (offset < 3)
       {sprintf(gszBuf, "CLIMB -0.%02d",gnClimbToneCps);lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    if (offset < 4)
       {sprintf(gszBuf, "HOUR   %02bu",gnHrs);lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    if (offset < 5)
       {sprintf(gszBuf, "MINUTE %02bu",gnMin);lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    if (offset < 6)
       {sprintf(gszBuf, "DATE   %02bu",gnDom);lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    if (offset < 7)
       {sprintf(gszBuf, "MONTH  %s",szMon[gnMon-1]);lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}   
    if (offset < 8)
       {sprintf(gszBuf, "YEAR   %u",(u16)(2000+gnYar));lcd_SetPos(n,8); lcd_PrintSz(gszBuf); n++;}
    }

void ui_InitRunMode(void)
   {
   lcd_ClearScreen();
   sns_Init();
   sns_Reset();
   sns_TriggerTemperatureSample();
   gnAcqState = 0;
   PCA0CN  |= 0x40;
   gnUIMode = UIMODE_RUN;
   }

void ui_InitSetMode(void)
   {
   gnParSel = 0;
   lcd_ClearScreen();
   sns_ReadSample();
   ui_DisplaySettings();
   gnUIMode = UIMODE_SET;
   }

void ui_InitFLogMode(void)
   {
   gnFLogRec = 0;
   ui_DisplayFLog();
   gnUIMode = UIMODE_FLOG;
   }


void ui_InitSleepMode(void)
   {
   lcd_ClearScreen();
   bat_Voltage();
   lcd_SetPos(1,0);
   sprintf(gszBuf,"BATTERY %bu.%buV",gnVBatx10/10,gnVBatx10%10);
   lcd_PrintSz(gszBuf);
   //gnBatLifeMin += (gnTimeTick+300)/6000;
   //lcd_SetPos(2,0);
   //sprintf(gszBuf,"TOT %uHR %uMIN",gnBatLifeMin/60, gnBatLifeMin%60);
   //lcd_PrintSz(gszBuf);
   tmr_WaitTicks(100);
   lcd_SetPos(4,0);
   lcd_PrintSz("STANDBY...");
   tmr_WaitTicks(200);
   gnCps = 0;
   ui_Sleep();
   }

void ui_Sleep(void)
  {
  lcd_WriteCmd(0x24); // power down LCD display
  OSCXCN    = 0x20; // external clock is CMOS clock on P0.3 from DS1337
  CLKSEL    = 0x01; // SYSCLK =  external clock
  OSCICN    = 0x00; // disable internal oscillator
  gnUIMode = UIMODE_SLEEP;
  }

void ui_SleepHandler(void)
   {
   if (gbBtnSPressed)
      {
      gbBtnSPressed = 0;
      gnTimeTick = 0;
      gnSecTick = 0;
      OSCICN    = 0x80; // enable internal oscillator (clk/8)
      CLKSEL = 0x00;  // SYSCLK = internal oscillator
      rtc_ReadData();
      lcd_Init();     // power up lcd display
      lcd_ClearScreen();
      ui_InitFLogMode();
      }
   }


void var_writeFlightLog(void)
   {
#ifdef FLOG_USE_SERIAL_EEPROM
   u8 cnt;
   u8 addr = FLOG_EEP_ADDR + gnFLogOldestRec*FLOG_REC_BYTES;
#endif
   gnFLBuf[0] = gnYar;
   gnFLBuf[1] = gnMon;
   gnFLBuf[2] = gnDom;
   gnFLBuf[3] = gnLaunchHrs;   
   gnFLBuf[4] = gnLaunchMin;
   gnFLBuf[5] = gnFlightHrs;   
   gnFLBuf[6] = gnFlightMin;
   gnFLBuf[7] = (u8)(gnLaunchAltM & 0xFF);   
   gnFLBuf[8] = (u8)((gnLaunchAltM>>8) & 0xFF);
   gnFLBuf[9] = (u8)(gnMaxAltM & 0xFF);   
   gnFLBuf[10] = (u8)((gnMaxAltM>>8) & 0xFF);   
   gnFLBuf[11] = (u8)(gnLZAltM & 0xFF);   
   gnFLBuf[12] = (u8)((gnLZAltM>>8) & 0xFF);
   gnFLBuf[13] = (u8)(gnMaxCps/10);   
   gnFLBuf[14] = (u8)(gnMinCps/10);
   gnFLBuf[15] = (u8)gnMaxTempC;
   gnFLBuf[16] = (u8)gnMinTempC;
#ifdef FLOG_USE_FLASH
   flash_Update(FLOG_PAGE_ADDR + gnFLogOldestRec*FLOG_REC_BYTES, gnFLBuf, FLOG_REC_BYTES);
#endif
#ifdef FLOG_USE_SERIAL_EEPROM
   for (cnt = 0; cnt < FLOG_REC_BYTES; cnt++) smb_wbyte(SMB_ID_24LC02,addr+cnt,gnFLBuf[cnt]);
#endif
   gnFLogOldestRec++;
   if (gnFLogOldestRec >= FLOG_MAX_FLIGHTS) {gnFLogOldestRec = 0;}
   nvd_WriteFLogOldestRec();
   }


// gnFLogOldestRec is the index of the oldest flight record in the circular buffer, the one
// which is free to write to
// gnFLogRec is the display logical index of the flight record, 0 being the most recent flight record, 
// FLOG_MAX_FLIGHTS-1 being the oldest flight record.
// ix is the corresponding physical index of the display record in the circular buffer

void ui_DisplayFLog(void)
   {
   s8 ix;
   lcd_ClearScreen();
   lcd_SetPos(0,0); sprintf(gszBuf,"     -%02bu-",gnFLogRec+1);lcd_PrintSz(gszBuf);
   ix = gnFLogOldestRec - gnFLogRec - 1;
   // wrap around the circular buffer if needed to find the index of the record
   if (ix < 0) {ix += FLOG_MAX_FLIGHTS;}
#ifdef FLOG_USE_FLASH
   flash_ReadBuf ((char*)gnFLBuf, (u16) (FLOG_PAGE_ADDR+ix*FLOG_REC_BYTES), FLOG_REC_BYTES);
#endif
#ifdef FLOG_USE_SERIAL_EEPROM
   smb_rbuf(SMB_ID_24LC02,gnFLBuf,FLOG_EEP_ADDR+ix*FLOG_REC_BYTES,FLOG_REC_BYTES);
#endif
   CLAMP(gnFLBuf[0],0,99);
   CLAMP(gnFLBuf[1],1,12);
   CLAMP(gnFLBuf[2],1,31);
   CLAMP(gnFLBuf[3],0,23);
   CLAMP(gnFLBuf[4],0,59);
   CLAMP(gnFLBuf[5],0,9);
   CLAMP(gnFLBuf[6],0,59);
   lcd_SetPos(1,0);sprintf(gszBuf," 20%02bu %s %02bu",gnFLBuf[0],szMon[gnFLBuf[1]-1],gnFLBuf[2]); lcd_PrintSz(gszBuf);
   lcd_SetPos(2,0);sprintf(gszBuf,"%02bu:%02bu   %02bu:%02bu",gnFLBuf[3],gnFLBuf[4],gnFLBuf[5],gnFLBuf[6]); lcd_PrintSz(gszBuf);
   lcd_SetPos(3,0);sprintf(gszBuf,"%04d +%02bu.%bu %bd",(s16)gnFLBuf[7] + (((s16)gnFLBuf[8])<<8), gnFLBuf[13]/10,gnFLBuf[13]%10,(s8)gnFLBuf[15]); lcd_PrintSz(gszBuf);
   lcd_SetPos(4,0);sprintf(gszBuf,"%04d -%02bu.%bu %bd",(s16)gnFLBuf[9] + (((s16)gnFLBuf[10])<<8),gnFLBuf[14]/10,gnFLBuf[14]%10,(s8)gnFLBuf[16]) ; lcd_PrintSz(gszBuf);
   lcd_SetPos(5,0);sprintf(gszBuf,"%04d",(s16)gnFLBuf[11] + (((s16)gnFLBuf[12])<<8) ); lcd_PrintSz(gszBuf);
   }


void bat_Voltage(void)
   {
   u32 nAccum;
   u8 cnt = 8; // average 8 samples
   
   ADC0CN = 0x80; // enable
   nAccum = 0;
   while (cnt--)
      {
      AD0BUSY = 1;
      while (!AD0INT); // wait for conversion
      nAccum += (u32)ADC0;
      }
   ADC0CN = 0; // shutdown
   // Vbatx10 = (acc+4)/8   * 2 * 300 /1024 
   nAccum += 4;
   nAccum *= 75;
   nAccum >>= 10;
   gnVBatx10 = (u8)((nAccum+5)/10);   
   }



#ifdef FLOG_USE_FLASH

// flash_WriteByte
// This routine writes a byte to the linear FLASH address <faddr>.
void flash_WriteByte (u16 faddr, char byte)
   {
   bit EA_SAVE = EA;                   // preserve EA
   char xdata * data pwrite;           // FLASH write pointer
   EA = 0;                             // disable interrupts
   // change clock speed to slow, then restore later
   VDM0CN = 0x80;                      // enable VDD monitor
   RSTSRC = 0x02;                      // enable VDD monitor as a reset source
   pwrite = (char xdata *) faddr;
   FLKEY  = 0xA5;                      // Key Sequence 1
   FLKEY  = 0xF1;                      // Key Sequence 2
   PSCTL |= 0x01;                      // PSWE = 1
   VDM0CN = 0x80;                      // enable VDD monitor
   RSTSRC = 0x02;                      // enable VDD monitor as a reset source
   *pwrite = byte;                     // write the byte
   PSCTL &= ~0x01;                     // PSWE = 0
   EA = EA_SAVE;                       // restore interrupts
   }

// flash_ReadByte
// This routine reads a byte from the linear FLASH address <faddr>.
u8 flash_ReadByte (u16 faddr)
   {
   bit EA_SAVE = EA;                   // preserve EA
   char code * data pread;             // FLASH read pointer
   u8 byte;
   EA = 0;                             // disable interrupts
   pread = (char code *) faddr;
   byte = *pread;                      // read the byte
   EA = EA_SAVE;                       // restore interrupts
   return byte;
   }

// flash_ErasePage
// This routine erases the FLASH page containing the linear FLASH address <addr>.
void flash_ErasePage (u16 addr)
   {
   bit EA_SAVE = EA;                   // preserve EA
   char xdata * data pwrite;           // FLASH write pointer
   EA = 0;                          // disable interrupts
   // change clock speed to slow, then restore later
   VDM0CN = 0x80;                      // enable VDD monitor
   RSTSRC = 0x02;                   // enable VDD monitor as a reset source
   pwrite = (char xdata *) addr;
   FLKEY  = 0xA5;                   // Key Sequence 1
   FLKEY  = 0xF1;                   // Key Sequence 2
   PSCTL |= 0x03;                   // PSWE = 1; PSEE = 1
   VDM0CN = 0x80;                      // enable VDD monitor
   RSTSRC = 0x02;                   // enable VDD monitor as a reset source
   *pwrite = 0;                     // initiate page erase
   PSCTL &= ~0x03;                  // PSWE = 0; PSEE = 0
   EA = EA_SAVE;                    // restore interrupts
   }


// flash_Clear
// This routine erases <nbytes> starting at flash address <faddr> 
// by performing a read-modify-write operation using gScratchBuf as
// a temporary holding area.  This function accepts <nbytes> up to
// <FLASH_PAGESIZE>.
// Also assumes that the clear will not cross a page boundary

void flash_Clear(u16 faddr, u16 nbytes)
   {
   u16 pageStart;           // first address in page containing <faddr>
   u16 pageEnd;             // last address in page containing <faddr>

   pageStart = faddr & ~(FLASH_PAGESIZE - 1);
   pageEnd = pageStart + FLASH_PAGESIZE - 1;
   
   // 1. Erase Scratch Buffer (set all bytes to 0xFF)
   flash_EraseScratchBuf();

   // 2. Copy bytes from first byte of page until faddr-1, to Scratch buffer
   flash_ReadBuf(gScratchBuf,pageStart,faddr-pageStart);

   // 3. Copy from (faddr+nbytes) until pageEnd, to Scratch Buffer
   flash_ReadBuf(gScratchBuf+(faddr - pageStart + nbytes),faddr+nbytes,pageEnd - faddr - nbytes + 1);
   
   // 4. Erase full flash page
   flash_ErasePage(pageStart);
   
   // 5. Copy Scratch buffer to flash page
   flash_WriteBuf(pageStart,gScratchBuf,FLASH_PAGESIZE);
   }


// flash_Update
// This routine replaces <nbytes> from <pBuf> at the FLASH address <faddr>.  
// This function calls flash_Clear() to erase all destination bytes 
// prior to copying the bytes from <pBuf>. 
// This function accepts <nbytes> up to FLASH_PAGESIZE.
void flash_Update(u16 faddr, char *pBuf, u16 nbytes)
   {
   flash_Clear(faddr, nbytes);
   flash_WriteBuf(faddr, pBuf, nbytes);
   }


// flash_WriteBuf
// This routine copies <nbytes> from <pBuf> to the linear FLASH address <faddr>.
void flash_WriteBuf(u16 faddr, char *pBuf, u16 nbytes)
   {
   u16 cnt;
   for (cnt = 0; cnt < nbytes; cnt++) {
      flash_WriteByte(faddr+cnt, *pBuf++);
      }
   }

// flash_ReadBuf
// This routine copies <nbytes> from the linear FLASH address <faddr> to <pBuf>.
char * flash_ReadBuf (char *pBuf, u16 faddr, u16 nbytes)
   {
   u16 cnt;
   for (cnt = 0; cnt < nbytes; cnt++) {
      *pBuf++ = flash_ReadByte(faddr+cnt);
      }
   return pBuf;
   }

void flash_EraseScratchBuf(void)
   {
   u16 cnt  = FLASH_PAGESIZE;
   while (cnt--) gScratchBuf[cnt] = 0xFF;
   }


// flash_Fill
// This routine fills the FLASH beginning at <faddr> with <length> bytes
// assumes the page has been erased already
void flash_Fill (u16 faddr, u16 length, u8 val)
   {
   u16 cnt;
   for (cnt = 0; cnt < length; cnt++) {
      flash_WriteByte(faddr+cnt, val);
      }
   }
#endif